// This is the render filter version of the D#DFilter
// It is not as up to date as the D3DFilter class though
// Beware of bugs!!!

#include <windows.h>
#include <streams.h>
#include <initguid.h>
#include <olectl.h>
#include <math.h>
#include "StereoRenderer.h"

#include <initguid.h>
#if (1100 > _MSC_VER)
#include <olectlid.h>
#else
#include <olectl.h>
#endif

// Forward declarations
void Msg(TCHAR *szFormat, ...);
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );
WNDCLASSEX wc;


// A structure for our custom vertex type. We added texture coordinates
struct CUSTOMVERTEX
{
    FLOAT  x, y, z; // The position
    FLOAT  tu, tv;   // The texture coordinates
};

// Our custom FVF, which describes our custom vertex structure
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_TEX1)


// Setup data

const AMOVIESETUP_MEDIATYPE sudPinTypes =
{
    &MEDIATYPE_Video,           // Major type
    &MEDIASUBTYPE_NULL          // Minor type
};

const AMOVIESETUP_PIN sudPins =
{
    L"Input",                   // Name of the pin
    FALSE,                      // Is pin rendered
    FALSE,                      // Is an output pin
    FALSE,                      // Ok for no pins
    FALSE,                      // Allowed many
    &CLSID_NULL,                // Connects to filter
    L"Output",                  // Connects to pin
    1,                          // Number of pin types
    &sudPinTypes                // Details for pins
};

const AMOVIESETUP_FILTER sudSampVid =
{
    &CLSID_StereoRenderer,      // Filter CLSID
    L"Stereo Video Renderer",   // Filter name
    MERIT_DO_NOT_USE,           // Filter merit
    1,                          // Number pins
    &sudPins                    // Pin details
};

// List of class IDs and creator functions for the class factory. This
// provides the link between the OLE entry point in the DLL and an object
// being created. The class factory will call the static CreateInstance

CFactoryTemplate g_Templates[] = {
    { L"Stereo Video Renderer"
    , &CLSID_StereoRenderer
    , CStereoRenderer::CreateInstance
    , NULL
    , &sudSampVid }
  ,
    { L"Quality Property Page"
    , &CLSID_SampleQuality
    , CQualityProperties::CreateInstance }
};
int g_cTemplates = sizeof(g_Templates) / sizeof(g_Templates[0]);


//
// CreateInstance
//
// This goes in the factory template table to create new filter instances
//
CUnknown * WINAPI CStereoRenderer::CreateInstance(LPUNKNOWN pUnk, HRESULT *phr)
{
    return new CStereoRenderer(NAME("Sample Video Renderer"),pUnk,phr);

} // CreateInstance


#pragma warning(disable:4355)

//
// Constructor
//
CStereoRenderer::CStereoRenderer(TCHAR *pName,
                               LPUNKNOWN pUnk,
                               HRESULT *phr) :
    CBaseVideoRenderer(CLSID_StereoRenderer,pName,pUnk,phr),
	m_lVidWidth(0), m_lVidHeight(0), m_lVidPitch(0),
	m_pTexLeftIn(NULL), m_pTexRightIn(NULL),
	m_pTexOut(NULL),
	m_ppfUnfitness_left(NULL), m_ppfUnfitness_right(NULL),
	m_ppiDisparity_left(NULL), m_ppiDisparity_right(NULL),
    m_pfFitness(NULL),
	m_ppiHidden_left(NULL),m_ppiHidden_right(NULL),
	m_ppfLeftPyramid(NULL), m_ppfRightPyramid(NULL),
	m_pTexDispLeft(NULL), m_pTexDispRight(NULL),
	m_pTexUnfLeft(NULL), m_pTexUnfRight(NULL),
	m_pTexHiddenLeft(NULL), m_pTexHiddenRight(NULL),
	m_ppSurfDispLeft(NULL), m_ppSurfDispRight(NULL),
	m_ppSurfUnfLeft(NULL), m_ppSurfUnfRight(NULL),
	m_ppSurfHiddenLeft(NULL), m_ppSurfHiddenRight(NULL),
	m_pTexFitness(NULL), m_ppSurfFitness(NULL)
{

	ASSERT(pName);
    ASSERT(phr);

	// create logfile
	m_logfile = new ofstream(LOGFILE_LOCATION);

////////////////////////////////////////////////////////////////////CREATE WINDOW
    // Register the window class
    WNDCLASSEX i_wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      TEXT("D3D Filter"), NULL };
	m_wc = wc = i_wc;

	if(!SUCCEEDED(RegisterClassEx( &m_wc )))
		*m_logfile << LOCATION << "register failed";
 
	 // Create the application's window
	 hWnd = CreateWindow( TEXT("D3D Filter"), TEXT("D3D Filter: Render window"),
                              WS_OVERLAPPED|WS_VISIBLE, 100, 100, 300, 300,
                              GetDesktopWindow(), NULL, NULL, NULL );
	 

	if(hWnd == NULL)
		*m_logfile << LOCATION << "hWnd is NULL. Error is: " << GetLastError() << ".\n";

////////////////////////////////////////////////////////////////////////////////////


	// INITIALIZE D3D device
	if( FAILED( this->InitD3D(hWnd) ) )
		*m_logfile << LOCATION << "initializing of d3d failed";	


	// Create the vertex buffer.
	if( FAILED( m_pd3dDevice->CreateVertexBuffer( 4*sizeof(CUSTOMVERTEX),
                                      0, D3DFVF_CUSTOMVERTEX,
                                      D3DPOOL_DEFAULT, &m_pVB, NULL ) ) )
    {
        Msg(TEXT("Could not create a vertex buffer!"));
    }

    // Fill the vertex buffer. We are setting the tu and tv texture
    // coordinates, which range from 0.0 to 1.0	
	CUSTOMVERTEX pVertices[]=
	{
		//  x      y     z    tu tv 
		{ -1.0f,  1.0f, 0.0f, 0, 1 },	//  lower right - red
		{ -1.0f, -1.0f, 0.0f, 0, 0 },	//  lower left - green
		{  1.0f,  1.0f, 0.0f, 1, 1 },	//  upper left - blue
		{  1.0f, -1.0f, 0.0f, 1, 0 },	//  upper right - white
	};

	void* gVertices;
    if ( FAILED(m_pVB->Lock( 0, sizeof(pVertices), (void**)&gVertices, 0 ) ) )
    {
        Msg(TEXT("Could not lock the vertex buffer!"));
    }
	memcpy(gVertices,pVertices,sizeof(pVertices));
	m_pVB->Unlock();

	// Create all the pixel and vertex shaders
	if (FAILED(CreateAllShaders()))
	{       
		Msg(TEXT("Could not create shaders!"));
	}

	// Create the font used to render text in the image
	const LOGFONT tempLogFont = { 12,			//Height
							4,					//Width 
							0,					//Escapement 
							0,					//Orientation 
							FW_REGULAR, 		//Weight
							NULL,				//Italic 
							NULL,				//Underline 
							NULL, 				//StrikeOut
							ANSI_CHARSET, 		//CharSet
							OUT_DEFAULT_PRECIS, //OutPrecision
							CLIP_DEFAULT_PRECIS,//ClipPrecision 
							DEFAULT_QUALITY, 	//Quality
							DEFAULT_PITCH,		//PitchAndFamily
							TEXT("Arial") };	//FaceName
	if (FAILED(D3DXCreateFontIndirect(m_pd3dDevice, &tempLogFont, &m_pFont)))
		*m_logfile <<"Could not create the font object";

	m_logfile->flush();


} // (Constructor)


//
// Destructor
//
CStereoRenderer::~CStereoRenderer()
{
    delete m_logfile;

	// Closing the window. Is it needed?
	MSG msg;
    ZeroMemory( &msg, sizeof(msg) );
    msg.message=WM_QUIT;
    TranslateMessage( &msg );
    DispatchMessage( &msg );
	
	// Remove textures
	RemoveTextures();

	// Remove shaders
	RemoveShaders();

    if( m_pVB != NULL )
        m_pVB->Release();

    if( m_pd3dDevice != NULL )
        m_pd3dDevice->Release();

    if( m_pD3D != NULL )
        m_pD3D->Release();

    UnregisterClass( TEXT("D3D Filter"), m_wc.hInstance );

} // (Destructor)



//
// GetPin
//
// We only support one input pin and it is numbered zero
//
/*
CBasePin *CStereoRenderer::GetPin(int n)
{
    ASSERT(n == 0);
    if (n != 0) {
        return NULL;
    }

    // Assign the input pin if not already done so

    return m_pInputPin;

} // GetPin
*/


//
// NonDelegatingQueryInterface
//
// Overriden to say what interfaces we support and where
//
STDMETHODIMP CStereoRenderer::NonDelegatingQueryInterface(REFIID riid,void **ppv)
{
    CheckPointer(ppv,E_POINTER);

    if (riid == IID_ISpecifyPropertyPages) {
        return GetInterface((ISpecifyPropertyPages *)this, ppv);

    } 

    return CBaseVideoRenderer::NonDelegatingQueryInterface(riid,ppv);

} // NonDelegatingQueryInterface


//
// GetPages
//
// Return the CLSIDs for the property pages we support
//
STDMETHODIMP CStereoRenderer::GetPages(CAUUID *pPages)
{
    CheckPointer(pPages,E_POINTER);

    pPages->cElems = 1;
    pPages->pElems = (GUID *) CoTaskMemAlloc(sizeof(GUID));
    if (pPages->pElems == NULL)
        return E_OUTOFMEMORY;

    pPages->pElems[0] = CLSID_SampleQuality;
    return NOERROR;

} // GetPages


//
// DoRenderSample
//
// Have the drawing object render the current image
//
HRESULT CStereoRenderer::DoRenderSample(IMediaSample *pIn)
{
	///////////////////////////////////////////////////////////////////
	// THE CODE FOR THE STEREO MATCHING 
	///////////////////////////////////////////////////////////////////

    //1 Move image to texture
	//2 Perform matching
	//3 Show result

	// Upload frame to texture
	HRESULT hr;
	if (FAILED(hr= UploadTexture(pIn)))
	{
		return hr;
	}

	// Render the texture in D3D

	//---------------------------------------------------------------------------
	if (FAILED(hr= Render5()))
	{
		Msg(TEXT("Det small"));
		return hr;
	}
	
	Render1(m_pTexDispLeft, m_pTexHiddenLeft);

	//Render(m_pTexDispLeft);
	//---------------------------------------------------------------------------

	return S_OK;
} // DoRenderSample

//-----------------------------------------------------------------------------
// UploadTexture: A sample has been delivered. Copy it to the textures
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::UploadTexture( IMediaSample * pSample )
{
    BYTE  *pBmpBuffer, *pTxtBuffer1, *pTxtBuffer2;     // Bitmap buffer, texture buffer
    LONG  lTxtPitch1, lTxtPitch2;                // Pitch of bitmap, texture
    
    // Get the video bitmap buffer
    pSample->GetPointer( &pBmpBuffer );

    // Lock the Textures
    D3DLOCKED_RECT d3dlr1;
    if (SUCCEEDED(m_pTexLeftIn->LockRect(0, &d3dlr1, 0, 0)))
	{
		D3DLOCKED_RECT d3dlr2;
		if (SUCCEEDED(m_pTexRightIn->LockRect(0, &d3dlr2, 0, 0)))
		{
    
			// Get the textures buffer & pitch
			pTxtBuffer1 = static_cast<byte *>(d3dlr1.pBits);
			lTxtPitch1 = d3dlr1.Pitch;
			pTxtBuffer2 = static_cast<byte *>(d3dlr2.pBits);
			lTxtPitch2 = d3dlr2.Pitch;
    
			// Copy the bits    

			for(int y = 0; y < m_lVidHeight; y++ ) {
				BYTE *pBmpBufferOld = pBmpBuffer;
				BYTE *pTxtBufferOld1 = pTxtBuffer1;
				BYTE *pTxtBufferOld2 = pTxtBuffer2;
				
				// The first half
				for (int x = 0; x < m_lVidWidth; x++) {
					pTxtBuffer1[0] = pBmpBuffer[0];
					pTxtBuffer1[1] = pBmpBuffer[1];
					pTxtBuffer1[2] = pBmpBuffer[2];
					pTxtBuffer1 += 4;
					pBmpBuffer += 3;
				}

				// The second half
				for (int y = 0; y < m_lVidWidth; y++) {
					pTxtBuffer2[0] = pBmpBuffer[0];
					pTxtBuffer2[1] = pBmpBuffer[1];
					pTxtBuffer2[2] = pBmpBuffer[2];
					pTxtBuffer2 += 4;
					pBmpBuffer += 3;
				}

				pBmpBuffer = pBmpBufferOld + m_lVidPitch;
				pTxtBuffer1 = pTxtBufferOld1 + lTxtPitch1;
				pTxtBuffer2 = pTxtBufferOld2 + lTxtPitch2;
			}
		}
		else
			return E_FAIL;
        
		// Unlock the Textures
		if (FAILED(m_pTexLeftIn->UnlockRect(0)))
			return E_FAIL;
	}
	else
		return E_FAIL;

    if (FAILED(m_pTexRightIn->UnlockRect(0)))
        return E_FAIL;
    
    return S_OK;
}


//-----------------------------------------------------------------------------
// InitTexture: Creates all pixel and vertex shaders
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::CreateAllShaders()
{

	D3DVERTEXELEMENT9 decl[] = 
	{
		{ 0, 0,  D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 }, // pos
		{ 0, 0,  D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_UV, D3DDECLUSAGE_TEXCOORD, 1 }, // tex
		D3DDECL_END()
	};


    TCHAR        strPath[512];                      // location of the shader file
    LPD3DXBUFFER pCode;                             // assembled shader code
	LPD3DXBUFFER pErrorMsgs;						// Error messages
	
	/////////////////////////////////////////////////////////////////////////
    // Create the vertex shaders.
	/////////////////////////////////////////////////////////////////////////
	// VS_offset1
	lstrcpy( strPath, VERTEX_SHADER_LOCATION );
	lstrcat(strPath, TEXT("VS_offset1.txt"));
	HRESULT hr;
	if( FAILED(hr = D3DXAssembleShaderFromFile( strPath, NULL, NULL, D3DXSHADER_DEBUG, &pCode, &pErrorMsgs )) ) {
		if(hr==D3DERR_INVALIDCALL)
			Msg(TEXT("1 Assemble vertex shader from file FAILED: D3DERR_INVALIDCALL"));
		else if(hr==D3DXERR_INVALIDDATA)
			Msg(TEXT("1 Assemble vertex shader from file FAILED: D3DXERR_INVALIDDATA"));
		else if(hr==E_OUTOFMEMORY)
			Msg(TEXT("1 Assemble vertex shader from file FAILED: E_OUTOFMEMORY"));
		else
			Msg(TEXT("1 Assemble vertex shader from file FAILED: This error is not possible..."));

		//Msg(TEXT("Assemble vertex shader from file FAILED"));// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();

		return hr;
	}
	else
	{
		m_pd3dDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &m_hVS_offset1 );
	}
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();
	// VS_offset2	
	lstrcpy( strPath, VERTEX_SHADER_LOCATION );
	lstrcat(strPath, TEXT("VS_offset2.txt"));
	if( FAILED(hr = D3DXAssembleShaderFromFile( strPath, NULL, NULL, D3DXSHADER_DEBUG, &pCode, &pErrorMsgs )) ) {
		if(hr==D3DERR_INVALIDCALL)
			Msg(TEXT("2 Assemble vertex shader from file FAILED: D3DERR_INVALIDCALL"));
		else if(hr==D3DXERR_INVALIDDATA)
			Msg(TEXT("2 Assemble vertex shader from file FAILED: D3DXERR_INVALIDDATA"));
		else if(hr==E_OUTOFMEMORY)
			Msg(TEXT("2 Assemble vertex shader from file FAILED: E_OUTOFMEMORY"));
		else
			Msg(TEXT("2 Assemble vertex shader from file FAILED: This error is not possible..."));

		//Msg(TEXT("Assemble vertex shader from file FAILED"));// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();

		return hr;
	}
	else
	{
		m_pd3dDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &m_hVS_offset2 );
	}
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();	
	// VS_offset3	
	lstrcpy( strPath, VERTEX_SHADER_LOCATION );
	lstrcat(strPath, TEXT("VS_offset3.txt"));
	if( FAILED(hr = D3DXAssembleShaderFromFile( strPath, NULL, NULL, D3DXSHADER_DEBUG, &pCode, &pErrorMsgs )) ) {
		if(hr==D3DERR_INVALIDCALL)
			Msg(TEXT("3 Assemble vertex shader from file FAILED: D3DERR_INVALIDCALL"));
		else if(hr==D3DXERR_INVALIDDATA)
			Msg(TEXT("3 Assemble vertex shader from file FAILED: D3DXERR_INVALIDDATA"));
		else if(hr==E_OUTOFMEMORY)
			Msg(TEXT("3 Assemble vertex shader from file FAILED: E_OUTOFMEMORY"));
		else
			Msg(TEXT("3 Assemble vertex shader from file FAILED: This error is not possible..."));

		//Msg(TEXT("Assemble vertex shader from file FAILED"));// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();

		return hr;
	}
	else
	{
		m_pd3dDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &m_hVS_offset3 );
	}
	// VS_offset4	
	lstrcpy( strPath, VERTEX_SHADER_LOCATION );
	lstrcat(strPath, TEXT("VS_offset4.txt"));
	if( FAILED(hr = D3DXAssembleShaderFromFile( strPath, NULL, NULL, D3DXSHADER_DEBUG, &pCode, &pErrorMsgs )) ) {
		if(hr==D3DERR_INVALIDCALL)
			Msg(TEXT("4 Assemble vertex shader from file FAILED: D3DERR_INVALIDCALL"));
		else if(hr==D3DXERR_INVALIDDATA)
			Msg(TEXT("4 Assemble vertex shader from file FAILED: D3DXERR_INVALIDDATA"));
		else if(hr==E_OUTOFMEMORY)
			Msg(TEXT("4 Assemble vertex shader from file FAILED: E_OUTOFMEMORY"));
		else
			Msg(TEXT("4 Assemble vertex shader from file FAILED: This error is not possible..."));

		//Msg(TEXT("Assemble vertex shader from file FAILED"));// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();

		return hr;
	}
	else
	{
		m_pd3dDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &m_hVS_offset4 );
	}
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();


	/////////////////////////////////////////////////////////////////////////
	// Create the pixel shaders
	/////////////////////////////////////////////////////////////////////////
	// BasicPixelShader
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("BasicPixelShader.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hBasicPixelShader );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	// PS_mark_hidden
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_mark_hidden.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_mark_hidden );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	// PS_difference
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_difference.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_difference );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	// PS_add4
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_add4.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_add4 );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	// PS_test_better
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_test_better.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_test_better );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();
	
	// PS_find_hidden
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_find_hidden.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_find_hidden );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	
	// PS_test_better
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	lstrcat(strPath, TEXT("PS_create_sample.txt"));
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
		
		return hr;
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hPS_create_sample );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();
   
	
	return S_OK;

}

//-----------------------------------------------------------------------------
// InitTexture: Creates all textures and surfaces 
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::InitTexture()
{
	HRESULT hr;
	
	///////////////////////////////////////////////////////////////////////////
	// Input textures
	///////////////////////////////////////////////////////////////////////////
	if( FAILED( hr = m_pd3dDevice->CreateTexture(
                    m_lVidWidth, m_lVidHeight,
                    0, D3DUSAGE_AUTOGENMIPMAP, 
                    D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexLeftIn, NULL ) ) )
    {
        Msg(TEXT("Could not create the D3DX texture!"));
        return hr;
    }
	if( FAILED( hr = m_pd3dDevice->CreateTexture(
                    m_lVidWidth, m_lVidHeight,
                    0, D3DUSAGE_AUTOGENMIPMAP, 
                    D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexRightIn, NULL ) ) )
    {
        Msg(TEXT("Could not create the D3DX texture!"));
        return hr;
    }
	if( FAILED( hr = m_pd3dDevice->CreateTexture(
                    m_lVidWidth, m_lVidHeight,
                    0, D3DUSAGE_AUTOGENMIPMAP, 
                    D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexOut, NULL ) ) )
    {
        Msg(TEXT("Could not create the D3DX texture!"));
        return hr;
    }


	///////////////////////////////////////////////////////////////////////////
	// Stuff used by the reference Renderers
	// Observe that the simple renderer and the SSD renderer only uses the lowest levels
	///////////////////////////////////////////////////////////////////////////
	m_pfFitness = new float[m_lVidWidth*m_lVidHeight];
	
	// Number of levels in pyramids
	m_iNbrLevels = 1;
	int width = m_lVidWidth;
	while (width>16)
	{
		width=width/2;
		m_iNbrLevels++;
	}

	m_ppfLeftPyramid = new float*[m_iNbrLevels];
	m_ppfRightPyramid = new float*[m_iNbrLevels];
	m_ppiDisparity_left = new int*[m_iNbrLevels];
	m_ppiDisparity_right	= new int*[m_iNbrLevels];	
	m_ppfUnfitness_left = new float*[m_iNbrLevels];
	m_ppfUnfitness_right	= new float*[m_iNbrLevels];
	m_ppiHidden_left = new int*[m_iNbrLevels];
	m_ppiHidden_right = new int*[m_iNbrLevels];


	width = m_lVidWidth;
	int height = m_lVidWidth;
	for (int i = 0;i<m_iNbrLevels;i++)
	{
		m_ppfLeftPyramid[i]=new float[width*height*3];
		m_ppfRightPyramid[i]=new float[width*height*3];	
		m_ppiDisparity_left[i] = new int[width*height];
		m_ppiDisparity_right[i]	= new int[width*height];	
		m_ppfUnfitness_left[i] = new float[width*height];
		m_ppfUnfitness_right[i]	= new float[width*height];
		m_ppiHidden_left[i] = new int[width*height];
		m_ppiHidden_right[i] = new int[width*height];
		width=width/2;
		height=height/2;
	}

	///////////////////////////////////////////////////////////////////////////
	// Stuff used by Render5()
	///////////////////////////////////////////////////////////////////////////
	
	// Create render target textures which are required to be a power of 2
	// D3DXCreateTexture takes care of this but the actual size is checked afterwards
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexDispLeft) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexDispRight) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexUnfLeft) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexUnfRight) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexHiddenLeft) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexHiddenRight ) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    m_iNbrLevels,D3DUSAGE_RENDERTARGET, 
                    D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pTexFitness ) ) )
    {
        Msg(TEXT("Could not create the D3DX render target texture!"));
        return hr;
    }
	// Check the actual size of the render targets
	D3DSURFACE_DESC Desc;
	if (FAILED(m_pTexDispLeft->GetLevelDesc(0,&Desc)))
		Msg(TEXT("Could not get the descption of the texture"));
	m_iTexWidth=Desc.Width;
	m_iTexHeight=Desc.Height;

	m_ppSurfDispLeft = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfDispRight = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfUnfLeft = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfUnfRight = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfHiddenLeft = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfHiddenRight = new LPDIRECT3DSURFACE9[m_iNbrLevels];		
	m_ppSurfFitness = new LPDIRECT3DSURFACE9[m_iNbrLevels];
	for (i = 0;i< m_iNbrLevels;i++)
	{
		UINT j =i;
		if( FAILED( hr = m_pTexDispLeft->GetSurfaceLevel(j, &m_ppSurfDispLeft[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}

		if( FAILED( hr = m_pTexDispRight->GetSurfaceLevel(j, &m_ppSurfDispRight[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}
		if( FAILED( hr = m_pTexUnfLeft->GetSurfaceLevel(j, &m_ppSurfUnfLeft[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}
		if( FAILED( hr = m_pTexUnfRight->GetSurfaceLevel(j, &m_ppSurfUnfRight[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}
		if( FAILED( hr = m_pTexHiddenLeft->GetSurfaceLevel(j, &m_ppSurfHiddenLeft[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}
		if( FAILED( hr = m_pTexHiddenRight->GetSurfaceLevel(j, &m_ppSurfHiddenRight[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}		
		if( FAILED( hr = m_pTexFitness->GetSurfaceLevel(j, &m_ppSurfFitness[i])))
		{
			Msg(TEXT("Could not get surface of render target texture!"));
			return hr;
		}
	}
	// Calculate the size of texels
	// Video textures and render target textures
	m_fTexelWidth=1.0f/m_lVidWidth;
	m_fTexelHeight=1.0f/m_lVidHeight;
	m_fHalfTexelWidth=0.5f/m_lVidWidth;
	m_fHalfTexelHeight=0.5f/m_lVidHeight;

	return S_OK;
}

//-----------------------------------------------------------------------------
// RemoveTexture: This method removes all textures and surfaces
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::RemoveTextures()
{
	//Input textures
	if( m_pTexLeftIn != NULL )
        m_pTexLeftIn->Release();
	if( m_pTexRightIn != NULL )
        m_pTexRightIn->Release();

	//Output textures
	if( m_pTexOut != NULL )
        m_pTexOut->Release();


	///////////////////////////////////////////////////////////////////////////
	//Reference renderers 
	///////////////////////////////////////////////////////////////////////////
	if (m_ppfLeftPyramid != NULL)
	{
		for (int i = 0;i<m_iNbrLevels;i++)
		{
			delete[] m_ppfLeftPyramid[i];
			delete[] m_ppfRightPyramid[i];
			delete[] m_ppfUnfitness_left[i];
			delete[] m_ppfUnfitness_right[i];
			delete[] m_ppiDisparity_left[i];
			delete[] m_ppiDisparity_right[i];
			delete[] m_ppiHidden_left[i];
			delete[] m_ppiHidden_right[i];
		}
		delete m_ppfLeftPyramid;
		delete m_ppfRightPyramid;
		delete[] m_ppfUnfitness_left;
		delete[] m_ppfUnfitness_right;
		delete[] m_ppiDisparity_left;
		delete[] m_ppiDisparity_right;
		delete[] m_ppiHidden_left;
		delete[] m_ppiHidden_right;
	}

	if( m_pfFitness != NULL )
	    delete[] m_pfFitness;

	///////////////////////////////////////////////////////////////////////////
	// Hardware renderers
	///////////////////////////////////////////////////////////////////////////
	for (int i = 0;i<m_iNbrLevels;i++)
	{
		if( m_ppSurfDispLeft[i] != NULL )
			m_ppSurfDispLeft[i]->Release();
		if( m_ppSurfDispRight[i] != NULL )
			m_ppSurfDispRight[i]->Release();
		if( m_ppSurfUnfLeft[i] != NULL )
			m_ppSurfUnfLeft[i]->Release();
		if( m_ppSurfUnfRight[i] != NULL )
			m_ppSurfUnfRight[i]->Release();
		if( m_ppSurfHiddenLeft[i] != NULL )
			m_ppSurfHiddenLeft[i]->Release();
		if( m_ppSurfHiddenRight[i] != NULL )
			m_ppSurfHiddenRight[i]->Release();
		if( m_ppSurfFitness[i] != NULL )
			m_ppSurfFitness[i]->Release();
	}
	
	if (m_ppSurfDispLeft!=NULL)
	{
		delete[] m_ppSurfDispLeft;
		delete[] m_ppSurfDispRight;
		delete[] m_ppSurfUnfLeft;
		delete[] m_ppSurfUnfRight;
		delete[] m_ppSurfHiddenLeft;
		delete[] m_ppSurfHiddenRight;
		delete[] m_ppSurfFitness;
	}

	// Mipmap textures
	if( m_pTexDispLeft != NULL )
        m_pTexDispLeft->Release();
	if( m_pTexDispRight != NULL )
        m_pTexDispRight->Release();
	if( m_pTexUnfLeft != NULL )
        m_pTexUnfLeft->Release();
	if( m_pTexUnfRight != NULL )
        m_pTexUnfRight->Release();
	if( m_pTexHiddenLeft != NULL )
        m_pTexHiddenLeft->Release();
	if( m_pTexHiddenRight != NULL )
        m_pTexHiddenRight->Release();
	if( m_pTexFitness != NULL )
        m_pTexFitness->Release();

	if( m_pBackBuffer != NULL)
		m_pBackBuffer->Release();

	return S_OK;
}


//-----------------------------------------------------------------------------
// RemoveShaders: This method removes all shaders
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::RemoveShaders()
{
	if( m_hVS_offset1 != NULL )
		m_hVS_offset1->Release();
	if( m_hVS_offset2 != NULL )
		m_hVS_offset2->Release();
	if( m_hVS_offset3 != NULL )
		m_hVS_offset3->Release();
	if( m_hVS_offset4 != NULL )
		m_hVS_offset4->Release();

	if( m_hBasicPixelShader != NULL )
		m_hBasicPixelShader->Release();
	if( m_hPS_mark_hidden != NULL )
		m_hPS_mark_hidden->Release();
	if( m_hPS_difference != NULL )
		m_hPS_difference->Release();	
	if( m_hPS_add4 != NULL )
		m_hPS_add4->Release();
	if( m_hPS_test_better != NULL )
		m_hPS_test_better->Release();
	if( m_hPS_create_sample != NULL )
		m_hPS_create_sample->Release();
	if( m_hPS_find_hidden != NULL )
		m_hPS_find_hidden->Release();
	
	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::InitD3D( HWND hWnd )
{

    // Create the D3D object.
    if( NULL == ( m_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE d3ddm;
    if( FAILED( m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = d3ddm.Format;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    // Create the D3DDevice
    if( FAILED( m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_MULTITHREADED,
                                      &d3dpp, &m_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    // Turn off culling
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

	// Texture wrapping
	m_pd3dDevice->SetSamplerState(0,D3DSAMP_ADDRESSU,D3DTADDRESS_BORDER);
	m_pd3dDevice->SetSamplerState(0,D3DSAMP_ADDRESSV,D3DTADDRESS_BORDER);
	m_pd3dDevice->SetSamplerState(0,D3DSAMP_ADDRESSW,D3DTADDRESS_BORDER);

	m_pd3dDevice->GetRenderTarget(0,&m_pBackBuffer);

    return S_OK;
}

//
// CheckMediaType
//
// Check the proposed video media type
//
HRESULT CStereoRenderer::CheckMediaType(const CMediaType *pmt)
{
    HRESULT   hr = E_FAIL;
    VIDEOINFO *pvi;
    
    // Reject the connection if this is not a video type
    if( *pmt->FormatType() != FORMAT_VideoInfo ) {
        return E_INVALIDARG;
    }

    // Retrive the size of this media type
    VIDEOINFO *pviBmp;                      // Bitmap info header
    pviBmp = (VIDEOINFO *)pmt->Format();
    m_lVidWidth  = pviBmp->bmiHeader.biWidth/2;
    m_lVidHeight = abs(pviBmp->bmiHeader.biHeight);

    // Only accept RGB24
    pvi = (VIDEOINFO *)pmt->Format();
    if(IsEqualGUID( *pmt->Type(),    MEDIATYPE_Video)  &&
       IsEqualGUID( *pmt->Subtype(), MEDIASUBTYPE_RGB24))
    {
        hr = S_OK;
    }
    
    return hr;

} // CheckMediaType

//
// SetMediaType
//
// We store a copy of the media type used for the connection in the renderer
// because it is required by many different parts of the running renderer
// This can be called when we come to draw a media sample that has a format
// change with it. We normally delay type changes until they are really due
// for rendering otherwise we will change types too early if the source has
// allocated a queue of samples. In our case this isn't a problem because we
// only ever receive one sample at a time so it's safe to change immediately
//
HRESULT CStereoRenderer::SetMediaType(const CMediaType *pmt)
{
//    HRESULT hr;

    // Retrive the size of this media type
    VIDEOINFO *pviBmp;                      // Bitmap info header
    pviBmp = (VIDEOINFO *)pmt->Format();
    m_lVidWidth  = pviBmp->bmiHeader.biWidth/2;
    m_lVidHeight = abs(pviBmp->bmiHeader.biHeight);
    m_lVidPitch = (pviBmp->bmiHeader.biWidth * 3 + 3) & ~(3); // We are forcing RGB24
	RemoveTextures();
	InitTexture();

    // Create the texture that maps to this media type
/*    if( FAILED( hr = D3DXCreateTexture(m_pd3dDevice,
                    m_lVidWidth, m_lVidHeight,
                    1, 0, 
                    D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexture ) ) )
    {
        //Msg(TEXT("Could not create the D3DX texture!  hr=0x%x"), hr);
        return hr;
    }

    // D3DXCreateTexture can silently change the parameters on us
    D3DSURFACE_DESC ddsd;
    if ( FAILED( hr = m_pTexture->GetLevelDesc( 0, &ddsd ) ) ) {
        //Msg(TEXT("Could not get level Description of D3DX texture! hr = 0x%x"), hr);
        return hr;
    }

  */
    return S_OK;
} // SetMediaType

//
// Active
//
// The auto show flag is used to have the window shown automatically when we
// change state. We do this only when moving to paused or running, when there
// is no outstanding EC_USERABORT set and when the window is not already up
// This can be changed through the IVideoWindow interface AutoShow property.
// If the window is not currently visible then we are showing it because of
// a state change to paused or running, in which case there is no point in
// the video window sending an EC_REPAINT as we're getting an image anyway
//
  //  HRESULT CStereoRenderer::Active()
  //  {
   //  HWND hwnd; = m_VideoText.GetWindowHWND();
  //     NOTE("AutoShowWindow");

  //      if(/*m_VideoText.IsAutoShowEnabled() ==*/ TRUE)
  //      {
  //          if(m_bAbort == FALSE)
  //          {
  //              if(/*IsWindowVisible(hwnd) == */FALSE)
  //              {
  //                  NOTE("Executing AutoShowWindow");
  //                  SetRepaintStatus(FALSE);

/*                m_VideoText.PerformanceAlignWindow();
                m_VideoText.DoShowWindow(SW_SHOWNORMAL);
                m_VideoText.DoSetWindowForeground(TRUE);*/
  //              }
  //          }
  //      }

  //      return CBaseVideoRenderer::Active();

  //  } // Active
//
// BreakConnect
//
// This is called when a connection or an attempted connection is terminated
// and lets us to reset the connection flag held by the base class renderer
// The filter object may be hanging onto an image to use for refreshing the
// video window so that must be freed (the allocator decommit may be waiting
// for that image to return before completing) then we must also uninstall
// any palette we were using, reset anything set with the control interfaces
// then set our overall state back to disconnected ready for the next time

  //  HRESULT CStereoRenderer::BreakConnect()
  //  {
  //     CAutoLock cInterfaceLock(&m_InterfaceLock);

    // Check we are in a valid state

  //      HRESULT hr = CBaseVideoRenderer::BreakConnect();
  //      if (FAILED(hr)) {
  //          return hr;
  //      }

    // The window is not used when disconnected
  /*  IPin *pPin = m_InputPin.GetConnected();
    if (pPin) 
        SendNotifyWindow(pPin,NULL);
*/
    // The base class break connect disables us from sending any EC_REPAINT
    // events which is important otherwise when we come down here to remove
    // our palette we end up painting the window again - which in turn sees
    // there is no image to draw and would otherwise send a redundant event

  //      m_mtIn.ResetFormatBuffer();

  //      return NOERROR;

  //  } // BreakConnect


//
// CompleteConnect
//
// When we complete connection we need to see if the video has changed sizes
// If it has then we activate the window and reset the source and destination
// rectangles. If the video is the same size then we bomb out early. By doing
// this we make sure that temporary disconnections such as when we go into a
// fullscreen mode do not cause unnecessary property changes. The basic ethos
// is that all properties should be persistent across connections if possible
//
  //  HRESULT CStereoRenderer::CompleteConnect(IPin *pReceivePin)
  //  {
  //      CAutoLock cInterfaceLock(&m_InterfaceLock);

  //      CBaseVideoRenderer::CompleteConnect(pReceivePin);
//    m_DrawImage.ResetPaletteVersion();

    // Has the video size changed between connections

  //      VIDEOINFOHEADER *pVideoInfo = (VIDEOINFOHEADER *) m_mtIn.Format();
  //      if (pVideoInfo->bmiHeader.biWidth == m_VideoSize.cx) 
  //      {
  //          if (pVideoInfo->bmiHeader.biHeight == m_VideoSize.cy) 
  //          {
  //              return NOERROR;
  //          }
  //      }

    // Pass the video window handle upstream
  /*  HWND hwnd;// = m_VideoText.GetWindowHWND();
    NOTE1("Sending EC_NOTIFY_WINDOW %x",hwnd);

    SendNotifyWindow(pReceivePin,hwnd);
*/
    // Set them for the current video dimensions

//    m_DrawImage.SetDrawContext();
  //      m_VideoSize.cx = pVideoInfo->bmiHeader.biWidth;
  //      m_VideoSize.cy = pVideoInfo->bmiHeader.biHeight;
/*
    m_VideoText.SetDefaultSourceRect();
    m_VideoText.SetDefaultTargetRect();
    m_VideoText.OnVideoSizeChange();
    m_VideoText.ActivateWindow();
*/
  //      return NOERROR;

  //  } // CompleteConnect


//
// OnReceiveFirstSample
//
// Use the image just delivered to display a poster frame
//
  //  void CStereoRenderer::OnReceiveFirstSample(IMediaSample *pMediaSample)
  //  {
  //      ASSERT(pMediaSample);
    
  //      DoRenderSample(pMediaSample);

  //  } // OnReceiveFirstSample


////////////////////////////////////////////////////////////////////////
//
// Exported entry points for registration and unregistration 
// (in this case they only call through to default implementations).
//
////////////////////////////////////////////////////////////////////////

//
// DllRegisterSever
//
// Handle the registration of this filter
//
STDAPI DllRegisterServer()
{
    return AMovieDllRegisterServer2( TRUE );

} // DllRegisterServer


//
// DllUnregisterServer
//
STDAPI DllUnregisterServer()
{
    return AMovieDllRegisterServer2( FALSE );

} // DllUnregisterServer


//
// DllEntryPoint
//
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

BOOL APIENTRY DllMain(HANDLE hModule, 
                      DWORD  dwReason, 
                      LPVOID lpReserved)
{
	return DllEntryPoint((HINSTANCE)(hModule), dwReason, lpReserved);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Msg: Display an error message box if needed
//-----------------------------------------------------------------------------
void Msg(TCHAR *szFormat, ...)
{
    //TCHAR szBuffer[512];

    //va_list pArgs;
    //va_start(pArgs, szFormat);
    //_vstprintf(szBuffer, szFormat, pArgs);
    //va_end(pArgs);

    MessageBox(NULL, szFormat, TEXT("DirectShow Texture3D Sample"), 
               MB_OK | MB_ICONERROR);
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

//-----------------------------------------------------------------------------
// round: Used to find the closest integer value
//-----------------------------------------------------------------------------
int round(double dIn)
{
	if (ceil(dIn)-0.5<0)
		return (int) ceil(dIn);
	else
		return (int) floor(dIn);

}

//-----------------------------------------------------------------------------
// Name: DownloadMatrix()
// Desc: Moves the reference disparity to output texture
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::DownloadMatrix()	// Moves the m_piDisparity_Left to m_pTexOut for rendering
{
	BYTE  *pTxtBuffer1;     // Bitmap buffer, texture buffer
    LONG  lTxtPitch1;

    // Lock the Textures
    D3DLOCKED_RECT d3dlr1;
    if (SUCCEEDED(m_pTexOut->LockRect(0, &d3dlr1, 0, 0)))
	{

		// Get the textures buffer & pitch
		pTxtBuffer1 = static_cast<byte *>(d3dlr1.pBits);
		lTxtPitch1 = d3dlr1.Pitch;

		// Copy the bits    
		for(int y = 0; y < m_lVidHeight; y++ ) {
			BYTE *pTexBufferOld = pTxtBuffer1;
			for (int x = 0; x < m_lVidWidth; x++) {
				if (!m_ppiHidden_left[0][y*m_lVidHeight+x])
				{
					pTxtBuffer1[0] = static_cast<unsigned char>(m_ppiDisparity_left[0][y*m_lVidHeight+x]*m_lVidWidth/SEARCH_LENGTH);
					pTxtBuffer1[1] = static_cast<unsigned char>(m_ppiDisparity_left[0][y*m_lVidHeight+x]*m_lVidWidth/SEARCH_LENGTH);
					pTxtBuffer1[2] = static_cast<unsigned char>(m_ppiDisparity_left[0][y*m_lVidHeight+x]*m_lVidWidth/SEARCH_LENGTH);
					pTxtBuffer1[3] = 0x00;
				}
				else
				{	
					pTxtBuffer1[0] = 0x00;
					pTxtBuffer1[1] = 0x00;
					pTxtBuffer1[2] = 0xAA;
					pTxtBuffer1[3] = 0x00;
				}
				pTxtBuffer1 += 4;
			}
			pTxtBuffer1 = pTexBufferOld + lTxtPitch1;
		}
    }
	else
	{
		*m_logfile << LOCATION << "download matrix failed\n";	
		m_logfile->flush();
		return E_FAIL;
	}

	
    if (FAILED(m_pTexOut->UnlockRect(0)))
	{
		*m_logfile << LOCATION << "download matrix failed in unlocking texture\n";	
		m_logfile->flush();
        return E_FAIL;
	}


    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------------------
// STEREO RENDERERS 
// WITH GROWING COMPLEXITY
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID CStereoRenderer::Render(LPDIRECT3DTEXTURE9 pTexOut)
{
	HRESULT hr;

	// To avoid the texture sliding, half a texel must be added
	float displacement[4] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
	if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement, 1)))                                      
    {
        Msg(TEXT("Could not set Vertex Shader displacement vector!"));
    }

	//Set the correct rendertarget (the backbuffer)
	m_pd3dDevice->SetRenderTarget(0,m_pBackBuffer);

	// Begin the scene
	if (SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		// Set texture
		m_pd3dDevice->SetTexture( 0, pTexOut);

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
		
		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hBasicPixelShader );
		m_pd3dDevice->SetVertexShader( m_hVS_offset1 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		//Write something onto the frame
		TCHAR pVarValue[40] ;
		lstrcpy( pVarValue, TEXT("D3D-filter has been here") );
		RECT tempRect = {15,15,900,100};
		if (FAILED(m_pFont->DrawText(pVarValue, -1, &tempRect, DT_LEFT, D3DCOLOR_ARGB(0,0,0,0))))
		{
			*m_logfile << LOCATION << "Failed to render the text onto the frame!";	
			m_logfile->flush();
		}

		// End scene
		m_pd3dDevice->EndScene();

		// Present
		m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
	}

}


//-----------------------------------------------------------------------------
// Name: Render1(LPDIRECT3DTEXTURE9 pTexOut, LPDIRECT3DTEXTURE9 pTexHidden)
// Desc: Draws the input texture on m_pRenderTargetSurface for output
//		 Uses the render to syrface object
//		 Compensates for hidden pixels
//-----------------------------------------------------------------------------
VOID CStereoRenderer::Render1(LPDIRECT3DTEXTURE9 pTexOut, LPDIRECT3DTEXTURE9 pTexHidden)
{
	HRESULT hr;

	// To avoid the texture sliding, half a texel must be added
	float displacement[8] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
								m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
	if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement, 2)))                                      
    {
        Msg(TEXT("Could not set Vertex Shader displacement vector!"));
    }

	//Set the correct rendertarget (the backbuffer)
	m_pd3dDevice->SetRenderTarget(0,m_pBackBuffer);

	// Begin the scene
	if (SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		// Set texture
		m_pd3dDevice->SetTexture( 0, pTexOut);
		m_pd3dDevice->SetTexture( 1, pTexHidden);

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
		
		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hPS_mark_hidden );
		m_pd3dDevice->SetVertexShader( m_hVS_offset2 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		//Write something onto the frame
		TCHAR pVarValue[40] ;
		lstrcpy( pVarValue, TEXT("D3D-filter has been here") );
		RECT tempRect = {15,15,900,100};
		if (FAILED(m_pFont->DrawText(pVarValue, -1, &tempRect, DT_LEFT, D3DCOLOR_ARGB(0,0,0,0))))
		{
			*m_logfile << LOCATION << "Failed to render the text onto the frame!";	
			m_logfile->flush();
		}

		// End scene
		m_pd3dDevice->EndScene();

		// Present
		m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
	}

}


//-----------------------------------------------------------------------------
// Name: Render2()
// Desc: Linear Reference Renderer
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::Render2()
{
 	BYTE  *pTex1Buffer, *pTex2Buffer;	// Pointers to the textures
	LONG  lTex1Pitch, lTex2Pitch;		// Pitch of the textures

	//	Lock input textures and get the textures buffer & pitch
    D3DLOCKED_RECT d3dlr1,d3dlr2;
    if (SUCCEEDED(m_pTexLeftIn->LockRect(0, &d3dlr1, 0, 0)))
	{
		if (SUCCEEDED(m_pTexRightIn->LockRect(0, &d3dlr2, 0, 0)))
		{
			pTex1Buffer = static_cast<byte *>(d3dlr1.pBits);
			lTex1Pitch = d3dlr1.Pitch;
			pTex2Buffer = static_cast<byte *>(d3dlr2.pBits);
			lTex2Pitch = d3dlr2.Pitch;

			//	Initialize the unfitness matrices

			for (int i = 0; i < m_lVidWidth*m_lVidHeight; i++)
			{
				m_ppfUnfitness_left[0][i]=MAX_FITNESS;
				m_ppfUnfitness_right[0][i]=MAX_FITNESS;
			}

			// For every displacement
			for (i=0; i<SEARCH_LENGTH ; i++)
			{
				// Reset the buffer pointers
				BYTE *pTex1BufferCopy = pTex1Buffer;
				BYTE *pTex2BufferCopy = pTex2Buffer;

				// Find current squares of differences 
				for (int k = 0; k<m_lVidHeight ; k++)
				{
					BYTE *pTex1BufferOld = pTex1BufferCopy; 
					BYTE *pTex2BufferOld = pTex2BufferCopy;
					pTex1BufferCopy += 4*i;
					for (int l = i; l<m_lVidWidth; l++)
					{
						float fTempB = static_cast<float>(pTex1BufferCopy[0])-static_cast<float>(pTex2BufferCopy[0]);
						float fTempG = static_cast<float>(pTex1BufferCopy[1])-static_cast<float>(pTex2BufferCopy[1]);
						float fTempR = static_cast<float>(pTex1BufferCopy[2])-static_cast<float>(pTex2BufferCopy[2]);			
						m_pfFitness[k*m_lVidWidth+l]=(fTempR*fTempR + fTempG*fTempG + fTempB*fTempB)/3;
						pTex1BufferCopy += 4;
						pTex2BufferCopy += 4;
					}
					pTex1BufferCopy=pTex1BufferOld + lTex1Pitch;
					pTex2BufferCopy=pTex2BufferOld + lTex2Pitch;
				}

				// Test the left fitness found and update if better
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						if (m_ppfUnfitness_left[0][k*m_lVidWidth+l]>m_pfFitness[k*m_lVidWidth+l])
						{
							m_ppfUnfitness_left[0][k*m_lVidWidth+l]=m_pfFitness[k*m_lVidWidth+l];
							m_ppiDisparity_left[0][k*m_lVidWidth+l]=i;
						}
					}
				}

				// Test the right fitness found and update if better
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						if (m_ppfUnfitness_right[0][k*m_lVidWidth+l-i]>m_pfFitness[k*m_lVidWidth+l])
						{
							m_ppfUnfitness_right[0][k*m_lVidWidth+l-i]=m_pfFitness[k*m_lVidWidth+l];
							m_ppiDisparity_right[0][k*m_lVidWidth+l-i]=i;
						}
					}
				}
			}
		}
		else
			Msg(TEXT("Could not lock the texture"));
		
		
		if (FAILED(m_pTexRightIn->UnlockRect(0)))
			Msg(TEXT("Could not unlock the texture"));
	}
	else
        Msg(TEXT("Could not lock the texture"));

	//	Unlock the texture
	if (FAILED(m_pTexLeftIn->UnlockRect(0)))
        Msg(TEXT("Could not unlock the texture"));

	// Find hidden areas
	// Initiate hidden
	for (int i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		m_ppiHidden_left[0][i]=0;
		m_ppiHidden_right[0][i]=0;
	}
	// Fill hidden with resampled disparities
	for (i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		int index1=i + m_ppiDisparity_right[0][i];
		int index2=i - m_ppiDisparity_left[0][i];
		if ((index1>=0)&&(index1<m_lVidWidth*m_lVidHeight))
			m_ppiHidden_left[0][index1]=m_ppiDisparity_right[0][i];
		if ((index2>=0)&&(index2<m_lVidWidth*m_lVidHeight))
			m_ppiHidden_right[0][index2]=m_ppiDisparity_left[0][i];
	}
	// Compare maps
	for (i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		m_ppiHidden_left[0][i]=(abs(m_ppiHidden_left[0][i] - m_ppiDisparity_left[0][i])>1); 
		m_ppiHidden_right[0][i]=(abs(m_ppiHidden_right[0][i] - m_ppiDisparity_right[0][i])>1); 
	}

	return DownloadMatrix();
}


//-----------------------------------------------------------------------------
// Name: Render3()
// Desc: SSD Reference Renderer using a 4x4 kernel
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::Render3()
{
 	BYTE  *pTex1Buffer, *pTex2Buffer;	// Pointers to the textures
	LONG  lTex1Pitch, lTex2Pitch;		// Pitch of the textures

	//	Lock input textures and get the textures buffer & pitch
    D3DLOCKED_RECT d3dlr1,d3dlr2;
    if (SUCCEEDED(m_pTexLeftIn->LockRect(0, &d3dlr1, 0, 0)))
	{
		if (SUCCEEDED(m_pTexRightIn->LockRect(0, &d3dlr2, 0, 0)))
		{
			pTex1Buffer = static_cast<byte *>(d3dlr1.pBits);
			lTex1Pitch = d3dlr1.Pitch;
			pTex2Buffer = static_cast<byte *>(d3dlr2.pBits);
			lTex2Pitch = d3dlr2.Pitch;

			//	Initialize the unfitness matrices

			for (int i = 0; i < m_lVidWidth*m_lVidHeight; i++)
			{
				m_ppfUnfitness_left[0][i]=MAX_FITNESS;
				m_ppfUnfitness_right[0][i]=MAX_FITNESS;
			}

			// For every displacement
			for (i=0; i<SEARCH_LENGTH ; i++)
			{
				// Reset the buffer pointers
				BYTE *pTex1BufferCopy = pTex1Buffer;
				BYTE *pTex2BufferCopy = pTex2Buffer;

				// Find current squares of differences 
				for (int k = 0; k<m_lVidHeight ; k++)
				{
					BYTE *pTex1BufferOld = pTex1BufferCopy; 
					BYTE *pTex2BufferOld = pTex2BufferCopy;
					pTex1BufferCopy += 4*i;
					for (int l = i; l<m_lVidWidth; l++)
					{
						float fTempB = static_cast<float>(pTex1BufferCopy[0])-static_cast<float>(pTex2BufferCopy[0]);
						float fTempG = static_cast<float>(pTex1BufferCopy[1])-static_cast<float>(pTex2BufferCopy[1]);
						float fTempR = static_cast<float>(pTex1BufferCopy[2])-static_cast<float>(pTex2BufferCopy[2]);			
						m_pfFitness[k*m_lVidWidth+l]=(fTempR*fTempR + fTempG*fTempG + fTempB*fTempB)/3;
						pTex1BufferCopy += 4;
						pTex2BufferCopy += 4;
					}
					pTex1BufferCopy=pTex1BufferOld + lTex1Pitch;
					pTex2BufferCopy=pTex2BufferOld + lTex2Pitch;
				}

				// Convolve the fitness map 1x1 -> 2x2
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth-1 ; l++)
					{
						m_pfFitness[k*m_lVidWidth+l] = (m_pfFitness[k*m_lVidWidth+l]+m_pfFitness[k*m_lVidWidth+l+1])/2;
					}
				}
				for (k = 1; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						m_pfFitness[k*m_lVidWidth+l] = (m_pfFitness[k*m_lVidWidth+l]+m_pfFitness[(k-1)*m_lVidWidth+l])/2;
					}
				}
				// Convolve the fitness map 2x2 -> 4x4
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i+2; l<m_lVidWidth ; l++)
					{
						m_pfFitness[k*m_lVidWidth+l] = (m_pfFitness[k*m_lVidWidth+l]+m_pfFitness[k*m_lVidWidth+l-2])/2;
					}
				}
				for (k = 0; k<m_lVidHeight-2 ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						m_pfFitness[k*m_lVidWidth+l] = (m_pfFitness[k*m_lVidWidth+l]+m_pfFitness[(k+2)*m_lVidWidth+l+1])/2;
					}
				}

				// Test the left fitness found and update if better
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						if (m_ppfUnfitness_left[0][k*m_lVidWidth+l]>m_pfFitness[k*m_lVidWidth+l])
						{
							m_ppfUnfitness_left[0][k*m_lVidWidth+l]=m_pfFitness[k*m_lVidWidth+l];
							m_ppiDisparity_left[0][k*m_lVidWidth+l]=i;
						}
					}
				}

				// Test the right fitness found and update if better
				for (k = 0; k<m_lVidHeight ; k++)
				{
					for (int l = i; l<m_lVidWidth ; l++)
					{
						if (m_ppfUnfitness_right[0][k*m_lVidWidth+l-i]>m_pfFitness[k*m_lVidWidth+l])
						{
							m_ppfUnfitness_right[0][k*m_lVidWidth+l-i]=m_pfFitness[k*m_lVidWidth+l];
							m_ppiDisparity_right[0][k*m_lVidWidth+l-i]=i;
						}
					}
				}
			}
		}
		else
			Msg(TEXT("Could not lock the texture"));
		
		
		if (FAILED(m_pTexRightIn->UnlockRect(0)))
			Msg(TEXT("Could not unlock the texture"));
	}
	else
        Msg(TEXT("Could not lock the texture"));

	//	Unlock the texture
	if (FAILED(m_pTexLeftIn->UnlockRect(0)))
        Msg(TEXT("Could not unlock the texture"));


	// Find hidden areas
	// Initiate hidden
	for (int i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		m_ppiHidden_left[0][i]=0;
		m_ppiHidden_right[0][i]=0;
	}
	// Fill hidden with resampled disparities
	for (i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		int index1=i + m_ppiDisparity_right[0][i];
		int index2=i - m_ppiDisparity_left[0][i];
		if ((index1>=0)&&(index1<m_lVidWidth*m_lVidHeight))
			m_ppiHidden_left[0][index1]=m_ppiDisparity_right[0][i];
		if ((index2>=0)&&(index2<m_lVidWidth*m_lVidHeight))
			m_ppiHidden_right[0][index2]=m_ppiDisparity_left[0][i];
	}
	// Compare maps
	for (i = 0;i<m_lVidWidth*m_lVidHeight;i++)
	{
		m_ppiHidden_left[0][i]=(abs(m_ppiHidden_left[0][i] - m_ppiDisparity_left[0][i])>1); 
		m_ppiHidden_right[0][i]=(abs(m_ppiHidden_right[0][i] - m_ppiDisparity_right[0][i])>1); 
	}

	return DownloadMatrix();
}


//-----------------------------------------------------------------------------
// Name: Render4()
// Desc: Hierarchical Reference Renderer which only matches the left disparity map
//			without finding occluded pixels
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::Render4()
{
 	BYTE  *pTex1Buffer, *pTex2Buffer;	// Pointers to the textures
	LONG  lTex1Pitch, lTex2Pitch;		// Pitch of the textures
	int iWidth = m_lVidWidth;
	int iHeight = m_lVidHeight;

//1Create pyramids

	// Lock input textures and get the textures buffer & pitch for first level of pyramid
	// This is no optimal solution
    D3DLOCKED_RECT d3dlr1,d3dlr2;
    if (SUCCEEDED(m_pTexLeftIn->LockRect(0, &d3dlr1, 0, 0)))
	{
		if (SUCCEEDED(m_pTexRightIn->LockRect(0, &d3dlr2, 0, 0)))
		{
			pTex1Buffer = static_cast<byte *>(d3dlr1.pBits);
			lTex1Pitch = d3dlr1.Pitch;
			pTex2Buffer = static_cast<byte *>(d3dlr2.pBits);
			lTex2Pitch = d3dlr2.Pitch;

			int iCurrent = 0;
			// Copy the images from the textures
			for (int k = 0; k<m_lVidHeight ; k++)
			{
				BYTE *pTex1BufferOld = pTex1Buffer; 
				BYTE *pTex2BufferOld = pTex2Buffer;
				for (int l = 0; l<m_lVidWidth; l++)
				{
					m_ppfLeftPyramid[0][iCurrent] = static_cast<float>(pTex1Buffer[0]);
					m_ppfRightPyramid[0][iCurrent] = static_cast<float>(pTex2Buffer[0]);
					m_ppfLeftPyramid[0][iCurrent+1] = static_cast<float>(pTex1Buffer[1]);
					m_ppfRightPyramid[0][iCurrent+1] = static_cast<float>(pTex2Buffer[1]);
					m_ppfLeftPyramid[0][iCurrent+2] = static_cast<float>(pTex1Buffer[2]);
					m_ppfRightPyramid[0][iCurrent+2] = static_cast<float>(pTex2Buffer[2]);			

					pTex1Buffer += 4;
					pTex2Buffer += 4;
					iCurrent += 3;
				}
				pTex1Buffer=pTex1BufferOld + lTex1Pitch;
				pTex2Buffer=pTex2BufferOld + lTex2Pitch;
			}

		}
		else
			Msg(TEXT("Could not lock the texture"));
		
		//	Unlock the texture
		if (FAILED(m_pTexRightIn->UnlockRect(0)))
			Msg(TEXT("Could not unlock the texture"));
	}
	else
        Msg(TEXT("Could not lock the texture"));
	
	//	Unlock the texture
	if (FAILED(m_pTexLeftIn->UnlockRect(0)))
        Msg(TEXT("Could not unlock the texture"));

	// Inizialize the rest of the pyramid
	for( int i = 1; i<m_iNbrLevels; i++)
	{
		int iCurrent = 0;
		int iPrevious = 0;
		int iWidthOld = iWidth;
		iWidth /= 2;
		iHeight /= 2;
		// Create this level, each pixel is the average of four pixels on the previous level
		for (int k = 0; k<iHeight ; k++)
		{
			for (int l = 0; l<iWidth; l++)
			{
				m_ppfLeftPyramid[i][iCurrent] = (m_ppfLeftPyramid[i-1][iPrevious] + m_ppfLeftPyramid[i-1][iPrevious+4] +
						m_ppfLeftPyramid[i-1][iPrevious+iWidthOld] + m_ppfLeftPyramid[i-1][iPrevious+iWidthOld+4])/4;
				m_ppfRightPyramid[i][iCurrent] = (m_ppfRightPyramid[i-1][iPrevious]+ m_ppfRightPyramid[i-1][iPrevious+4] +
						m_ppfRightPyramid[i-1][iPrevious+iWidthOld] + m_ppfRightPyramid[i-1][iPrevious+iWidthOld+4])/4;
				m_ppfLeftPyramid[i][iCurrent+1] = (m_ppfLeftPyramid[i-1][iPrevious+1]+ m_ppfLeftPyramid[i-1][iPrevious+5] +
						m_ppfLeftPyramid[i-1][iPrevious+iWidthOld+1] + m_ppfLeftPyramid[i-1][iPrevious+iWidthOld+5])/4;
				m_ppfRightPyramid[i][iCurrent+1] = (m_ppfRightPyramid[i-1][iPrevious+1]+ m_ppfRightPyramid[i-1][iPrevious+4] +
						m_ppfRightPyramid[i-1][iPrevious+iWidthOld+1] + m_ppfRightPyramid[i-1][iPrevious+iWidthOld+5])/4;
				m_ppfLeftPyramid[i][iCurrent+2] = (m_ppfLeftPyramid[i-1][iPrevious+2]+ m_ppfLeftPyramid[i-1][iPrevious+6] +
						m_ppfLeftPyramid[i-1][iPrevious+iWidthOld+2] + m_ppfLeftPyramid[i-1][iPrevious+iWidthOld+6])/4;
				m_ppfRightPyramid[i][iCurrent+2] = (m_ppfRightPyramid[i-1][iPrevious+2]+ m_ppfRightPyramid[i-1][iPrevious+4] +
						m_ppfRightPyramid[i-1][iPrevious+iWidthOld+2] + m_ppfRightPyramid[i-1][iPrevious+iWidthOld+6])/4;

				iCurrent+=3;
				iPrevious+=6;
			}
			iPrevious+=iWidthOld;
		}
	}


//2Match upper level
	int iCurrentLevel=m_iNbrLevels-1;
	// Initilize to differece
	for (i=0;i<iWidth*iHeight;i++)
	{
		m_ppfUnfitness_left[m_iNbrLevels-1][i]=float( pow(m_ppfLeftPyramid[iCurrentLevel][i*3]
												- m_ppfLeftPyramid[iCurrentLevel][i*3],2)
												+ pow(m_ppfLeftPyramid[iCurrentLevel][i*3+1]
												- m_ppfLeftPyramid[iCurrentLevel][i*3+1],2)
												+ pow(m_ppfLeftPyramid[iCurrentLevel][i*3+2]
												- m_ppfLeftPyramid[iCurrentLevel][i*3+2],2));
		m_ppfUnfitness_right[iCurrentLevel][i]=m_ppfUnfitness_left[iCurrentLevel][i];
		m_ppiDisparity_left[iCurrentLevel][i]=0;
		m_ppiDisparity_right[iCurrentLevel][i]=0;
	}
	// Offset one, test difference
	int iOffset=1;
	for (i=0;i<iHeight;i++)
	{
		for (int j=iOffset;j<iWidth;j++)
		{
			int iCurrent=i*iWidth+j;
			m_pfFitness[i*iWidth+j]=float( pow(m_ppfLeftPyramid[iCurrentLevel][iCurrent*3]
											- m_ppfLeftPyramid[iCurrentLevel][iCurrent*3-iOffset*3],2)
											+ pow(m_ppfLeftPyramid[iCurrentLevel][iCurrent*3+1]
											- m_ppfLeftPyramid[iCurrentLevel][iCurrent*3+1-iOffset*3],2)
											+ pow(m_ppfLeftPyramid[iCurrentLevel][iCurrent*3+2]
											- m_ppfLeftPyramid[iCurrentLevel][iCurrent*3+2-iOffset*3],2));
			// if better update disparity
			if (m_pfFitness[iCurrent]<m_ppfUnfitness_left[iCurrentLevel][iCurrent])
			{
				m_ppfUnfitness_left[iCurrentLevel][iCurrent]=m_pfFitness[iCurrent];
				m_ppiDisparity_left[iCurrentLevel][iCurrent]=iOffset;
			}

			if (m_pfFitness[iCurrent]<m_ppfUnfitness_left[iCurrentLevel][iCurrent-iOffset])
			{
				m_ppfUnfitness_right[iCurrentLevel][iCurrent-iOffset]=m_pfFitness[iCurrent];
				m_ppiDisparity_right[iCurrentLevel][iCurrent-iOffset]=iOffset;
			}
		}
	}
	

	int* piTempDisp = new int[m_lVidWidth*m_lVidHeight];


//For each level do
	while (iCurrentLevel>0)
	{


// Get and interpolate the input disparities
		//1 copy
		for (i = 0;i<iHeight;i++)
		{
			for (int j = 0;j<iWidth;j++)
			{
				int iCurrent = i*iWidth*2+j*2;
				int iPrevious = i*iWidth+j;
				m_ppiDisparity_left[iCurrentLevel-1][iCurrent] = m_ppiDisparity_left[iCurrentLevel][iPrevious];
				m_ppiDisparity_left[iCurrentLevel-1][iCurrent+1] = m_ppiDisparity_left[iCurrentLevel][iPrevious];
				m_ppiDisparity_left[iCurrentLevel-1][iCurrent+2*iWidth] = m_ppiDisparity_left[iCurrentLevel][iPrevious];
				m_ppiDisparity_left[iCurrentLevel-1][iCurrent+2*iWidth+1] = m_ppiDisparity_left[iCurrentLevel][iPrevious];
			}
		}

		//2 update
		iCurrentLevel--;
		iWidth*=2;
		iHeight*=2;
		int iPos;
		//3 convolve first middle
		for (i = 1;i<iHeight-1;i++)
		{
			for (int j = 1;j<iWidth-1;j++)
			{
				iPos = i*iWidth+j;
				piTempDisp[iPos]=round(2*(m_ppiDisparity_left[iCurrentLevel][iPos]
										+m_ppiDisparity_left[iCurrentLevel][iPos-1]
										+m_ppiDisparity_left[iCurrentLevel][iPos+1]
										+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth]
										+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth+1]
										+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth-1]
										+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth]
										+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth+1]
										+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth-1])/9);
			}
		}
		//3 convolve now the edges
		iPos = (iHeight-1)*iWidth;
		for (int j = 1;j<iWidth-1;j++)
		{
			piTempDisp[j]=round(2*(m_ppiDisparity_left[iCurrentLevel][j]
									+m_ppiDisparity_left[iCurrentLevel][j-1]
									+m_ppiDisparity_left[iCurrentLevel][j+1]
									+m_ppiDisparity_left[iCurrentLevel][j+iWidth]
									+m_ppiDisparity_left[iCurrentLevel][j+iWidth+1]
									+m_ppiDisparity_left[iCurrentLevel][j+iWidth-1])/6);

			piTempDisp[iPos+j]=round(2*(m_ppiDisparity_left[iCurrentLevel][iPos+j]
									+m_ppiDisparity_left[iCurrentLevel][iPos+j-1]
									+m_ppiDisparity_left[iCurrentLevel][iPos+j+1]
									+m_ppiDisparity_left[iCurrentLevel][iPos+j-iWidth]
									+m_ppiDisparity_left[iCurrentLevel][iPos+j-iWidth+1]
									+m_ppiDisparity_left[iCurrentLevel][iPos+j-iWidth-1])/6);
		}
		for (int i = 1;j<iHeight-1;j++)
		{
			iPos = i*iWidth;
			piTempDisp[iPos+j]=round(2*(m_ppiDisparity_left[iCurrentLevel][iPos]
									+m_ppiDisparity_left[iCurrentLevel][iPos+1]
									+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth]
									+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth+1]
									+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth]
									+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth+1])/6);
		
			int iPos = iPos+iWidth-1;
			piTempDisp[iPos+j]=round(2*(m_ppiDisparity_left[iCurrentLevel][iPos]
									+m_ppiDisparity_left[iCurrentLevel][iPos-1]
									+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth]
									+m_ppiDisparity_left[iCurrentLevel][iPos-iWidth-1]
									+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth]
									+m_ppiDisparity_left[iCurrentLevel][iPos+iWidth-1])/6);
		}
		//3 convolve now the corners
		piTempDisp[0]=2*m_ppiDisparity_left[iCurrentLevel][0];
		piTempDisp[iWidth]=2*m_ppiDisparity_left[iCurrentLevel][iPos];
		piTempDisp[iPos]=2*m_ppiDisparity_left[iCurrentLevel][iPos];
		piTempDisp[iPos+iWidth-1]=2*m_ppiDisparity_left[iCurrentLevel][iPos+iWidth-1];

// Match with modified SSD
		//Copy piTempDisp
		for (i=0;i<iHeight*iWidth;i++);
			m_ppiDisparity_left[iCurrentLevel][i]=piTempDisp[i];

		//Find initial unfitness map
		//FOR interval
		//Displace piTempDisp furter away
		//Check the displacements
		//Update initial unfitness and disparity
		//END
		
/*
fitness =   (left{N}(sub2ind([height,width],y_minus,x_plus))- right{N}(sub2ind([height,width], y_minus, pos +ones(height,width)))).^2 +...
            (left{N}(sub2ind([height,width],y,x_plus))- right{N}(sub2ind([height,width], y, pos +ones(height,width)))).^2 +...
            (left{N}(sub2ind([height,width],y_plus,x_plus))- right{N}(sub2ind([height,width], y_plus, pos +ones(height,width)))).^2+...
            (left{N}(sub2ind([height,width],y_minus,x))- right{N}(sub2ind([height,width], y_minus, pos ))).^2 +...
            (left{N}(sub2ind([height,width],y,x))- right{N}(sub2ind([height,width], y, pos ))).^2 +...
            (left{N}(sub2ind([height,width],y_plus,x))- right{N}(sub2ind([height,width], y_plus, pos ))).^2+...
            (left{N}(sub2ind([height,width],y_minus,x_minus))- right{N}(sub2ind([height,width], y_minus, pos -ones(height,width)))).^2 +...
            (left{N}(sub2ind([height,width],y,x_minus))- right{N}(sub2ind([height,width], y, pos -ones(height,width)))).^2 +...
            (left{N}(sub2ind([height,width],y_plus,x_minus))- right{N}(sub2ind([height,width], y_plus, pos -ones(height,width)))).^2;
%Test this fitness
mask1= fitness<unfitness_left{N};    
mask1=((x-pos)>=0).*mask1;
unfitness_left{N}=(~mask1).*unfitness_left{N} + mask1.*fitness;
disparity_left{N}=(~mask1).*disparity_left{N} + mask1.*(x-pos); 		
*/
		for (i = 1;i<iHeight-1;i++)
		{
			for (int j = 1;j<iWidth-1;j++)
			{
				m_pfFitness[i*iWidth+j] = float(
					pow( m_ppfLeftPyramid[iCurrentLevel][i*iWidth + j] - m_ppfRightPyramid[iCurrentLevel][i*iWidth + j + piTempDisp[i*iWidth+j]],2)
					);
			}
		}

	}
//End For

	delete piTempDisp;


// The resulting disparity in m_piDisparity_left


	return DownloadMatrix();
}

//-----------------------------------------------------------------------------
// Name: Render5()
// Desc: Simple linear Renderer version 1.1
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::Render5()
{
	HRESULT hr;

	/////////////////////////////////////////////////////////////////////////////////////
	//	Initialize the unfitness matrices (RGB - disparity, A - unfitness)
	/////////////////////////////////////////////////////////////////////////////////////
	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispLeft[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(255,255,255,255), 1.0f, 0 ))) {
		*m_logfile << LOCATION << "Failed to clear d3ddevice: " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispRight[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(255,255,255,255), 1.0f, 0 ))) {
		*m_logfile << LOCATION << "Failed to clear d3ddevice: " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// For every displacement
	/////////////////////////////////////////////////////////////////////////////////////
	for (int i=0; i<SEARCH_LENGTH ; i++)
	{
	
		/////////////////////////////////////////////////////////////////////////////////////
		// 1 Find current squares of difference (left disparity) output to fitness map		
		/////////////////////////////////////////////////////////////////////////////////////
		// Fitness Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Offset the right texture and left texture with vertex shader constants
		//
		// PS_difference is the Pixel Shader
		/////////////////////////////////////////////////////////////////////////////////////
		
		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Set offset for VS_offset2 as vertex shader constant
		// To avoid the textures sliding, half a texel must be added to both textures
		float fOffset = i*m_fTexelWidth;
		float displacement1[8] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									m_fHalfTexelWidth-fOffset,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement1, 2)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexLeftIn );
			m_pd3dDevice->SetTexture( 1, m_pTexRightIn );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );

			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );			
			m_pd3dDevice->SetPixelShader( m_hPS_difference );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// 2 Left test
		// Test the fitness found and update if better
		// output as left disp/fitness 
		/////////////////////////////////////////////////////////////////////////////////////
		// Left Disparity Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Fitness texture and Left Disparity texture are offset equal
		//
		// PS_test_better is the Pixel Shader
		//	  Current displacement is stored in a pixel shader constant
		/////////////////////////////////////////////////////////////////////////////////////

		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispLeft[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Set offset for VS_offset2 as vertex shader constant
		// To avoid the textures sliding, half a texel must be added to both textures
		float displacement2[4] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(1, displacement2, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}
		
		// Store current displacement which is used in PS_test_better
		float disparity[4] = {fOffset, fOffset, fOffset, fOffset};
		if (FAILED(hr =m_pd3dDevice->SetPixelShaderConstantF(0, disparity, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexDispLeft );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetPixelShader( m_hPS_test_better );
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );
			
			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// 3 Right test
		// Test fitness found and update if better (displace back to the right) 
		// output as right disp/fitness 
		/////////////////////////////////////////////////////////////////////////////////////
		// Right Disparity Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Fitness texture are offset to the right
		//
		// PS_test_better is the Pixel Shader
		//	  Current displacement is stored in a pixel shader constant
		/////////////////////////////////////////////////////////////////////////////////////

		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispRight[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		float displacement3[4] = { m_fHalfTexelWidth+fOffset,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement3, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexDispRight );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetPixelShader( m_hPS_test_better );
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}
	}

		
	/////////////////////////////////////////////////////////////////////////////////////
	// 4 Find hidden areas
	// Resample one texture according to itself and compare to the other
	//
	// 1 Create sampling texture from right disparity map
	/////////////////////////////////////////////////////////////////////////////////////
	// Fitness Surface is temporary render target
	// 
	// VS_offset1 is Vertex Shader
	//    Textures are offset to avoid sliding
	//
	// PS_create_sample  is the Pixel Shader
	//	  The input disparity is diplaced according to itself to create a sampling texture
	/////////////////////////////////////////////////////////////////////////////////////

	// Find hidden areas
	// Resample one texture according to itself
	// 1 Create sampling texture from right disparity map

	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	// Offset the right texture
	// To avoid the textures sliding, half a texel must be added to both
	float displacement4[12] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
								m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
								m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
	if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement4, 1)))                                      
	{
		Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
		return hr;
	}

	// Begin the scene
	if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
	{

		// Set textures
		m_pd3dDevice->SetTexture( 0, m_pTexDispLeft  );
		m_pd3dDevice->SetTexture( 1, m_pTexDispRight  );

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );

		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hPS_create_sample );
		m_pd3dDevice->SetVertexShader( m_hVS_offset2 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		// End the scene
		m_pd3dDevice->EndScene();
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// 4 Find hidden areas
	// Resample one texture according to itself and compare to the other
	//
	// 2 Resample right disparity map and compare to the left
	/////////////////////////////////////////////////////////////////////////////////////
	// Hidden Surface is render target
	// 
	// VS_offset3 is Vertex Shader
	//    Textures are offset to avoid sliding
	//
	// PS_find_hidden is the Pixel Shader
	//	  One disparity is offset according to itself and compared to the other
	/////////////////////////////////////////////////////////////////////////////////////

	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfHiddenLeft[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}

	// Begin the scene
	if SUCCEEDED( m_pd3dDevice->BeginScene())
	{

		// Set textures
		m_pd3dDevice->SetTexture( 0, m_pTexDispLeft  );
		m_pd3dDevice->SetTexture( 1, m_pTexFitness  );
		m_pd3dDevice->SetTexture( 2, m_pTexDispRight  );

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hPS_find_hidden );
		m_pd3dDevice->SetVertexShader( m_hVS_offset3 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		// End the scene
		m_pd3dDevice->EndScene();
	}


	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: Render6()
// Desc: SSD Renderer version 1.1 //Easy when 5 is ready
//-----------------------------------------------------------------------------
HRESULT CStereoRenderer::Render6()
{
	HRESULT hr;

	/////////////////////////////////////////////////////////////////////////////////////
	//	Initialize the unfitness matrices (RGB - disparity, A - unfitness)
	/////////////////////////////////////////////////////////////////////////////////////
	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispLeft[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(255,255,255,255), 1.0f, 0 ))) {
		*m_logfile << LOCATION << "Failed to clear d3ddevice: " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispRight[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	if(FAILED(hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(255,255,255,255), 1.0f, 0 ))) {
		*m_logfile << LOCATION << "Failed to clear d3ddevice: " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// For every displacement
	/////////////////////////////////////////////////////////////////////////////////////
	for (int i=0; i<SEARCH_LENGTH ; i++)
	{
	
		/////////////////////////////////////////////////////////////////////////////////////
		// 1 Find current squares of difference (left disparity) 
		// output to fitness map		
		/////////////////////////////////////////////////////////////////////////////////////
		// Fitness Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Offset the right texture and left texture with vertex shader constants
		//
		// PS_difference is the Pixel Shader
		/////////////////////////////////////////////////////////////////////////////////////
		
		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Set offset for VS_offset2 as vertex shader constant
		// To avoid the textures sliding, half a texel must be added to both textures
		float fOffset = i*m_fTexelWidth;
		float displacement1[8] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									m_fHalfTexelWidth-fOffset,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement1, 2)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexLeftIn );
			m_pd3dDevice->SetTexture( 1, m_pTexRightIn );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );

			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );			
			m_pd3dDevice->SetPixelShader( m_hPS_difference );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}


		/////////////////////////////////////////////////////////////////////////////////////
		// 2 Convolve current squares of difference (left disparity)
		// adding 4 pixels 2 times resulting in convolution with a 4x4 kernel
		// output to fitness map	
		//
		// 1 Add 4 (kernel 2x2) -> Add 16 (4x4)
		/////////////////////////////////////////////////////////////////////////////////////
		// Fitness Surface is render target
		// 
		// VS_offset4 is Vertex Shader
		//    Offset the textures to avoid sliding
		//
		// PS_add4 is the Pixel Shader
		/////////////////////////////////////////////////////////////////////////////////////
		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Offset to avoid sliding in VS_offset4
		float displacement5[16] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									3*m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									3*m_fHalfTexelWidth,3*m_fHalfTexelHeight,0.0f,0.0f,
									m_fHalfTexelWidth,3*m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement5, 4)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexFitness );
			m_pd3dDevice->SetTexture( 2, m_pTexFitness );
			m_pd3dDevice->SetTexture( 3, m_pTexFitness );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );

			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set Shaders
			m_pd3dDevice->SetPixelShader( m_hPS_add4 );
			m_pd3dDevice->SetVertexShader( m_hVS_offset4 );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// 2 Convolve current squares of difference (left disparity)
		// adding 4 pixels 2 times resulting in convolution with a 4x4 kernel
		// output to fitness map	
		//
		// 2 Add 16 (kernel 4x4)
		/////////////////////////////////////////////////////////////////////////////////////
		// Fitness Surface is render target
		// 
		// VS_offset4 is Vertex Shader
		//    Offset the textures to avoid sliding
		//
		// PS_add4 is the Pixel Shader
		/////////////////////////////////////////////////////////////////////////////////////
		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Offset the right texture
		// To avoid the textures sliding, half a texel must be added to both
		float displacement6[16] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									-2*m_fTexelWidth+ m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
									-2*m_fTexelWidth+ m_fHalfTexelWidth,-2*m_fTexelHeight+ m_fHalfTexelHeight,0.0f,0.0f,
									m_fHalfTexelWidth,-2*m_fTexelHeight+ m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement6, 4)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexFitness );
			m_pd3dDevice->SetTexture( 2, m_pTexFitness );
			m_pd3dDevice->SetTexture( 3, m_pTexFitness );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set Shaders
			m_pd3dDevice->SetPixelShader( m_hPS_add4 );
			m_pd3dDevice->SetVertexShader( m_hVS_offset4 );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}







		/////////////////////////////////////////////////////////////////////////////////////
		// 2 Left test
		// Test the fitness found and update if better
		// output as left disp/fitness 
		/////////////////////////////////////////////////////////////////////////////////////
		// Left Disparity Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Fitness texture and Left Disparity texture are offset equal
		//
		// PS_test_better is the Pixel Shader
		//	  Current displacement is stored in a pixel shader constant
		/////////////////////////////////////////////////////////////////////////////////////

		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispLeft[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		// Set offset for VS_offset2 as vertex shader constant
		// To avoid the textures sliding, half a texel must be added to both textures
		float displacement2[4] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(1, displacement2, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}
		
		// Store current displacement which is used in PS_test_better
		float disparity[4] = {fOffset, fOffset, fOffset, fOffset};
		if (FAILED(hr =m_pd3dDevice->SetPixelShaderConstantF(0, disparity, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexDispLeft );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetPixelShader( m_hPS_test_better );
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );
			
			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// 3 Right test
		// Test fitness found and update if better (displace back to the right) 
		// output as right disp/fitness 
		/////////////////////////////////////////////////////////////////////////////////////
		// Right Disparity Surface is render target
		// 
		// VS_offset2 is Vertex Shader
		//    Fitness texture are offset to the right
		//
		// PS_test_better is the Pixel Shader
		//	  Current displacement is stored in a pixel shader constant
		/////////////////////////////////////////////////////////////////////////////////////

		if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfDispRight[0]))) {
			*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
			m_logfile->flush();
			return hr;
		}

		float displacement3[4] = { m_fHalfTexelWidth+fOffset,m_fHalfTexelHeight,0.0f,0.0f};
		if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement3, 1)))                                      
		{
			Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
			return hr;
		}

		// Begin the scene
		if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			// Set textures
			m_pd3dDevice->SetTexture( 0, m_pTexFitness );
			m_pd3dDevice->SetTexture( 1, m_pTexDispRight );

			// Set vertex buffer stream source
			m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
			// Set current vertex stream format
			m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
			
			// Set shaders
			m_pd3dDevice->SetPixelShader( m_hPS_test_better );
			m_pd3dDevice->SetVertexShader( m_hVS_offset2 );

			// Render the vertex stream
			if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
				if(hr==D3DERR_INVALIDCALL) {
					*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
				else {
					*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
					m_logfile->flush();
				}
			}

			// End the scene
			m_pd3dDevice->EndScene();
		}
	}

		
	/////////////////////////////////////////////////////////////////////////////////////
	// 4 Find hidden areas
	// Resample one texture according to itself and compare to the other
	//
	// 1 Create sampling texture from right disparity map
	/////////////////////////////////////////////////////////////////////////////////////
	// Fitness Surface is temporary render target
	// 
	// VS_offset1 is Vertex Shader
	//    Textures are offset to avoid sliding
	//
	// PS_create_sample  is the Pixel Shader
	//	  The input disparity is diplaced according to itself to create a sampling texture
	/////////////////////////////////////////////////////////////////////////////////////

	// Find hidden areas
	// Resample one texture according to itself
	// 1 Create sampling texture from right disparity map

	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfFitness[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}
	// Offset the right texture
	// To avoid the textures sliding, half a texel must be added to both
	float displacement4[12] = { m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
								m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f,
								m_fHalfTexelWidth,m_fHalfTexelHeight,0.0f,0.0f};
	if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement4, 1)))                                      
	{
		Msg(TEXT("Could not set Vertex Shader displacement vectors!"));
		return hr;
	}

	// Begin the scene
	if ( SUCCEEDED(m_pd3dDevice->BeginScene()))
	{

		// Set textures
		m_pd3dDevice->SetTexture( 0, m_pTexDispLeft  );
		m_pd3dDevice->SetTexture( 1, m_pTexDispRight  );

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );

		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hPS_create_sample );
		m_pd3dDevice->SetVertexShader( m_hVS_offset2 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		// End the scene
		m_pd3dDevice->EndScene();
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// 4 Find hidden areas
	// Resample one texture according to itself and compare to the other
	//
	// 2 Resample right disparity map and compare to the left
	/////////////////////////////////////////////////////////////////////////////////////
	// Hidden Surface is render target
	// 
	// VS_offset3 is Vertex Shader
	//    Textures are offset to avoid sliding
	//
	// PS_find_hidden is the Pixel Shader
	//	  One disparity is offset according to itself and compared to the other
	/////////////////////////////////////////////////////////////////////////////////////

	if(FAILED(hr = m_pd3dDevice->SetRenderTarget(0, m_ppSurfHiddenLeft[0]))) {
		*m_logfile << LOCATION << "Failed to set render target " << (hr&0xffff) << endl;	
		m_logfile->flush();
		return hr;
	}

	// Begin the scene
	if SUCCEEDED( m_pd3dDevice->BeginScene())
	{

		// Set textures
		m_pd3dDevice->SetTexture( 0, m_pTexDispLeft  );
		m_pd3dDevice->SetTexture( 1, m_pTexFitness  );
		m_pd3dDevice->SetTexture( 2, m_pTexDispRight  );

		// Set vertex buffer stream source
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
			
		// Set current vertex stream format
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		// Set shaders
		m_pd3dDevice->SetPixelShader( m_hPS_find_hidden );
		m_pd3dDevice->SetVertexShader( m_hVS_offset3 );

		// Render the vertex stream
		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		// End the scene
		m_pd3dDevice->EndScene();
	}


	return S_OK;

}

