//-----------------------------------------------------------------------------
// Stereo Video Renderer
// Desc:
// A render filter which incorporates the same functions as the D3Dfilter
//
// Dependencies: 
// dxguid.lib strmbasd.lib msvcrtd.lib d3dx9d.lib d3d9.lib quartz.lib vfw32.lib winmm.lib version.lib largeint.lib comctl32.lib olepro32.lib
//-----------------------------------------------------------------------------


#include <d3dx9.h>
#include <d3dx9shader.h>
#include <d3dx9tex.h>
#include <fstream.h>
#include "vidprop.h"



// GUID
DEFINE_GUID(CLSID_StereoRenderer,
0x4d4b1600, 0x33ac, 0x11cf, 0xbf, 0x30, 0x00, 0xaa, 0x00, 0x55, 0x59, 0x5a);

// For now, but the shaders hav eto be copied!!!
#define VERTEX_SHADER_LOCATION TEXT("E:\\common\\StereoRenderer\\")
#define PIXEL_SHADER_LOCATION TEXT("E:\\common\\StereoRenderer\\")

#define LOGFILE_LOCATION "E:\\common\\cStereoRenderer_logfile.txt"
// This define is used for writing out the current file and line when logging
#define LOCATION "\n" << __FILE__ << " : Line "<< __LINE__ << "\n"

#define MAX_FITNESS 65536
#define SEARCH_LENGTH 128

#ifndef __STEREO_RENDERER__
#define __STEREO_RENDERER__


//-----------------------------------------------------------------------------
// Define GUID for Texture Renderer
// {71771540-2017-11cf-AE26-0020AFD79767}
//-----------------------------------------------------------------------------
//struct __declspec(uuid("{71771540-2017-11cf-ae26-0020afd79767}")) CLSID_TextureRenderer;

class CStereoRenderer : public ISpecifyPropertyPages, public CBaseVideoRenderer
{
public:

    // Constructor and destructor
    static CUnknown * WINAPI CreateInstance(LPUNKNOWN, HRESULT *);

    // Implement the ISpecifyPropertyPages interface

    DECLARE_IUNKNOWN
    STDMETHODIMP NonDelegatingQueryInterface(REFIID, void **);
    STDMETHODIMP GetPages(CAUUID *pPages);

    //CBasePin *GetPin(int n);

    // Override these from the filter and renderer classes	
	// These methods are required
	HRESULT CheckMediaType(const CMediaType *pmt );     // Format acceptable?
    HRESULT SetMediaType(const CMediaType *pmt );       // Video format notification
    HRESULT DoRenderSample(IMediaSample *pMediaSample); // New video sample

	// These methods are optional
   /* void PrepareRender();
    HRESULT Active();
    HRESULT BreakConnect();
    HRESULT CompleteConnect(IPin *pReceivePin);
    HRESULT CheckVideoType(const VIDEOINFO *pDisplay,const VIDEOINFO *pInput);
    HRESULT UpdateFormat(VIDEOINFO *pVideoInfo);
    void OnReceiveFirstSample(IMediaSample *pMediaSample);*/

private:

    // Constructor
    CStereoRenderer(TCHAR *pName,LPUNKNOWN pUnk,HRESULT *phr);
    ~CStereoRenderer();

	// D3D help methods
	HRESULT InitD3D( HWND hWnd );
	HRESULT InitTexture();
	HRESULT RemoveTextures();
	HRESULT CreateAllShaders();
	HRESULT RemoveShaders();
	HRESULT UploadTexture( IMediaSample * pIn );
	HRESULT DownloadMatrix();	// Moves the m_piDisparity_Left to m_pTexOut for rendering
	
	// Renderers
	VOID	Render(LPDIRECT3DTEXTURE9 pTexOut);									// Standard renderer of a texture
	VOID	Render1(LPDIRECT3DTEXTURE9 pTexOut, LPDIRECT3DTEXTURE9 pTexHidden);	// Compensates for hidden pixels
	HRESULT Render2();		// Linear reference renderer
	HRESULT Render3();		// SSD reference renderer
	HRESULT Render4();		// Hierarchical reference renderer
	HRESULT Render5();		// Linear Renderer version 1.1
	HRESULT Render6();		// SSD Renderer version 1.1

	// D3D objects
	LPDIRECT3D9				m_pD3D;	
	LPDIRECT3DDEVICE9		m_pd3dDevice;			// d3d device
	LPDIRECT3DVERTEXBUFFER9 m_pVB;					// Buffer to hold vertices
	WNDCLASSEX				m_wc;					// Needed for HWND
	HWND					hWnd;					// Hidden device window (for D3D)
	LPD3DXFONT				m_pFont;				// Font for drawing text
	LPDIRECT3DSURFACE9		m_pBackBuffer;			// Surfaces of the backbuffer

	// Textures
	LPDIRECT3DTEXTURE9      m_pTexLeftIn;				// Our input textures
	LPDIRECT3DTEXTURE9      m_pTexRightIn;				// Our input textures
	LPDIRECT3DTEXTURE9      m_pTexOut;					// Our output textures

	LPDIRECT3DTEXTURE9      m_pTexDispLeft;				// Disparity texture left	(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexDispRight;			// Disparity texture right	(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexUnfLeft;				// Unfitness texture left	(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexUnfRight;				// Unfitness texture right	(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexHiddenLeft;			// Hidden texture left		(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexHiddenRight;			// Hidden texture right		(Pyramid when used)
	LPDIRECT3DTEXTURE9      m_pTexFitness;					// Fitness texture			(Pyramid when used)

	LPDIRECT3DSURFACE9*		m_ppSurfDispLeft;			// Surfaces of the left disparity texture
	LPDIRECT3DSURFACE9*		m_ppSurfDispRight;			// Surfaces of the right disparity texture
	LPDIRECT3DSURFACE9*		m_ppSurfUnfLeft;			// Surfaces of the left unfitness texture
	LPDIRECT3DSURFACE9*		m_ppSurfUnfRight;			// Surfaces of the right unfitness texture
	LPDIRECT3DSURFACE9*		m_ppSurfHiddenLeft;			// Surfaces of the left hidden texture
	LPDIRECT3DSURFACE9*		m_ppSurfHiddenRight;		// Surfaces of the right hidden texture
	LPDIRECT3DSURFACE9*		m_ppSurfFitness;			// Surfaces of the right hidden texture

	float*					m_pfFitness;								// "Surface" (int[][]) used by Render2()
	
	int**					m_ppiDisparity_left;						// "Surfaces" (int[][]) (pyramids)
	int**					m_ppiDisparity_right;						// "Surfaces" (int[][]) (pyramids)			
	float**					m_ppfUnfitness_left; 						// "Surfaces" (float[][]) (pyramids)
	float**					m_ppfUnfitness_right;						// "Surfaces" (float[][]) (pyramids)
	int**					m_ppiHidden_left;							// "Surfaces" (int[][]) (pyramids)
	int**					m_ppiHidden_right;							// "Surface" (int[][]) (pyramids)
	float**					m_ppfLeftPyramid;							// "Scale-space" used by Render4()
	float**					m_ppfRightPyramid;							// "Scale-space" used by Render4()

	// Shaders
	LPDIRECT3DVERTEXSHADER9 m_hVS_offset1;			// Vertex shader handle
	LPDIRECT3DVERTEXSHADER9 m_hVS_offset2;			// Vertex shader handle
	LPDIRECT3DVERTEXSHADER9 m_hVS_offset3;			// Vertex shader handle
	LPDIRECT3DVERTEXSHADER9 m_hVS_offset4;			// Vertex shader handle


	LPDIRECT3DPIXELSHADER9 m_hBasicPixelShader;				// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_mark_hidden;				// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_add4;						// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_difference;				// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_test_better;				// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_create_sample;				// Pixel shader handle
	LPDIRECT3DPIXELSHADER9 m_hPS_find_hidden;				// Pixel shader handle

	// Video properties
	LONG m_lVidWidth;	// Video width
	LONG m_lVidHeight;	// Video Height
	LONG m_lVidPitch;	// Video Pitch
	float m_fTexelWidth;
	float m_fTexelHeight;
	float m_fHalfTexelWidth; // half a texel in width
	float m_fHalfTexelHeight; // half a texel in height

	// Scale-space properties
	int m_iNbrLevels;	// Number of pyramid levels required at the current video size
	int m_iTexWidth;
	int m_iTexHeight;


	// Log file
	ofstream *m_logfile;



}; // CVideoRenderer

#endif // __STEREO_RENDERER__
